/*
	solution for git sha based on
	http://stackoverflow.com/questions/1435953/how-can-i-pass-git-sha1-to-compiler-as-definition-using-cmake
*/

#ifndef __git_sha1_h_
#define __git_sha1_h_
extern const char g_GIT_SHA1[];
extern const char g_GIT_SHALONG[];
extern const char g_GIT_LAST_COMMIT_DATE[];

#endif
